//========================================================================
// File: bigdominic.cc
// Description: Large scale testing of dominic routing on Delaunay 
//              triangulations
// Author: Pat Morin
// Created: Mon Feb  1 19:02:05 EST 1999
//========================================================================
#include <LEDA/list.h>
#include <LEDA/point.h>
#include <LEDA/random_source.h>
#include <LEDA/plane_alg.h>
#include <LEDA/graph_alg.h>
#include <LEDA/window.h>

#include "dominic.h"

#ifndef min // return the minimum 
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max // return the maximum
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif


int main(int argc, char *argv[])
{
  // read command line arguments
  if (argc != 3) {
    cerr << "Usage: " << argv[0] << " <npoints> <nreps>" << endl;
    exit(1);
  }
  int npoints = atoi(argv[1]);
  int nreps = atoi(argv[2]);

  // create our persistent data structures
  GRAPH<point,int> T;
  random_source R;
  list<point> L;

  // maxratio measures the worst competitive ratio
  double maxratio = 1.0;

  for (int round = 0; round < nreps; round++) {
    // generate random points in square and triangulate them
    L.clear();
    for (int i = 0; i < npoints; i++) {
      double x, y;
      R >> x;
      R >> y;
      L.append(point(x,y));
    }
    T.clear();
    DELAUNAY_TRIANG(L, T);

    // setup the positions array for DOMINIC_PATH
    node_array<point> pos(T);
    node v;
    forall_nodes(v, T)
      pos[v] = T[v];
   
    // setup the cost array for SHORTEST_PATH
    edge_array<double> c(T);
    edge e;
    forall_edges (e, T)
      c[e] = T[T.source(e)].distance(T[T.target(e)]);
 
    // these hold the output of DOMINIC_PATH and SHORTEST_PATH
    node_array<edge> pred(T);
    node_array<double> cdist(T);
    node_array<double> sdist(T);

    // compute the competitive ratio of dominic routing on T
    node t;
    forall_nodes (t, T) {
      DOMINIC_PATH(T, t, pos, cdist, pred);
      SHORTEST_PATH(T, t, c, sdist, pred);
      node u;
      forall_nodes (u, T) {
	if (cdist[u] / sdist[u] > maxratio)
	  maxratio = cdist[u] / sdist[u]; 
      }
    }
  }
  // output the results
  cout << npoints <<  " " << maxratio << " (" << nreps << ")" << endl;

  return 0;
}





