//========================================================================
// File: comparison.c
// Description: A comparison of four online routing strategies.
// Author: Pat Morin
// Created: Wed Aug 19 13:20:25 EDT 1998
//========================================================================
#include <fstream.h>
#include <strstream.h>
#include <stdio.h>

#include <LEDA/list.h>
#include <LEDA/point.h>
#include <LEDA/random_source.h>
#include <LEDA/plane_alg.h>
#include <LEDA/graph_alg.h>
#include <LEDA/window.h>

#include "greedy.h"
#include "compass.h"
#include "rcompass.h"
#include "rhr.h"

// return the minimum of a and b
#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

// return the maximum of a and b
#ifndef max
#define max(a, b) ((a) > (b) ? (a) : (b))
#endif

// Run forever, generating random Delaunay triangulations and testing
// whether the conjecture described above is true for each triangulation.
int main(int argc, char *argv[])
{
  typedef void (*router)(graph&, node_array<point>&, node_matrix<double>&);

  router fns[4] = 
  { ALL_PAIRS_GREEDY_PATHS, ALL_PAIRS_COMPASS_PATHS, 
    ALL_PAIRS_RANDOMIZED_COMPASS_PATHS, ALL_PAIRS_RIGHTHAND_PATHS }; 

  char *names[] = 
  { "greedy", "compass", "rcompass", "rhr" };

  ofstream os[4];

  double hi[4], lo[4];

  // read command line arguments
  if (argc != 3 && argc != 4) {
    cerr << "Usage: " << argv[0] << " <npoints> <nreps> [d|g]" << endl;
    exit(1);
  }
  int npoints = atoi(argv[1]);
  int nreps = atoi(argv[2]);
  bool graham = (argc == 4) && (argv[3][0] == 'g');

  // initialize our arrays
  for (int i = 0; i < 4; i++) {
    hi[i] = 1.0;
    lo[i] = HUGE_VAL;
  }

  // create our persistent data structures
  GRAPH<point,int> T;
  random_source R;
  list<point> L;

  for (int round = 0; round < nreps; round++) {
    // generate random points in square
    L.clear();
    for (int i = 0; i < npoints; i++) {
      double x, y;
      R >> x;
      R >> y;
      L.append(point(x,y));
    }

    // compute their triangulation
    T.clear();
    if (graham) {
      TRIANGULATE_POINTS(L, T);
    } else {
      DELAUNAY_TRIANG(L, T);
    }

    // setup the positions array
    node_array<point> pos(T);
    node v;
    forall_nodes(v, T) {
      pos[v] = T[v];
    }

    // compute shortest paths
    node_matrix<double> sdist(T);
    edge_array<double> costs(T);
    edge e;
    forall_edges (e, T)
      costs[e] = T[T.source(e)].distance(T[T.target(e)]);
    ALL_PAIRS_SHORTEST_PATHS(T, costs, sdist);

    for (int i = 0; i < 4; i++) {
      // compute paths
      node_matrix<double> cdist(T);
      fns[i](T, pos, cdist);

      // measure the competitive ratio
      node u;
      double ratio = 1.0;
      forall_nodes (u, T) {
	forall_nodes (v, T) {
	  if (u != v) {
	    if (cdist[u][v] / sdist[u][v] > ratio)
	      ratio = cdist[u][v] / sdist[u][v];
	  }
	}
      }
      hi[i] = max(hi[i], ratio);
      lo[i] = min(lo[i], ratio);
    }
  }

  // output the results 
  for (int i = 0; i < 4; i++) {
    char buf[50];
    sprintf(buf, "%c-%s-%d", graham ? 'g' : 'd', names[i], npoints);
    os[i].open(buf);
    os[i] << npoints << " " << hi[i] << endl;
    os[i].close();
  }
  return 0;
}


