//==========================================================================
// File: dominic.c
// Description: An implementation of dominic routing
// Author: Pat Morin
// Created: Wed Jan 20 20:13:58 EST 1999
// Notes: If the packet is stored at vertex u and its destination is
// vertex v, the dominic routing strategy always moves the packet to
// the vertex w such that dist(u,w)+dist(w,v) is minimized.  This
// strategy isn't described in the paper, because it doesn't even work
// for Delaunay triangulations - trapping edges are really common.
//==========================================================================
#include "dominic.h"

#include <LEDA/list.h>

//==========================================================================
// Return the dominic successor of u when the destination is v
// Complexity: O(degree(u))
//========================================================================
inline 
edge dominic_successor(graph &G, node_array<point> &pos, node u, node v) 
{
  double dmin = HUGE_VAL;
  edge emin = nil;
  edge e;
  forall_adj_edges(e, u) {
    double d = pos[G.source(e)].distance(pos[G.target(e)]) 
      + pos[G.target(e)].distance(pos[v]); 
    if (d < dmin) {
      dmin = d;
      emin = e;
    }
  }
  return emin;
}

//==========================================================================
// Find the dominic routing path between all pairs of vertices in G.
// Complexity: O(n^2)
// Implementation: 
// Just uses DOMINIC_PATH n times.
//========================================================================
void ALL_PAIRS_DOMINIC_PATHS(graph &G, node_array<point>& pos, 
			    node_matrix<double>& dist)
{
  node_array<double> adist(G);
  node_array<edge> pred(G);

  node v;
  forall_nodes (v, G) {
    DOMINIC_PATH(G, v, pos, adist, pred);
    node u;
    forall_nodes (u, G)
      dist[u][v] = adist[u];
  }
}


//==========================================================================
// Find the dominic routing path from all vertices of G to t
// Complexity: O(n)
// Implementation:
// Memoizes subpaths in order to achieve linear running time.  dist[u]
// is set to -1 if u has not yet been visited, is set to 0 if u has
// been visited but its distance hasn't yet been computed, and is set
// to its correct value otherwise.  This is how we detect trapping
// cycles.
//========================================================================
void DOMINIC_PATH(graph &G, node t, node_array<point> &pos, 
		  node_array<double> &dist, node_array<edge> &pred)
{
  // initialize the distance array
  node u;
  forall_nodes (u, G) {
    dist[u] = -1.0;
  }
  dist[t] = 0.0;

  // initialize the only entry we need in the pred array
  pred[t] = nil;

  // find the dominic path from each node u to t
  forall_nodes (u, G) {
    node cur = u; // the current node
    list<node> P; // the entire dominic path from u to v
    P.push(cur);

    // keep going until we get to a node we've already visited
    while (dist[cur] < 0.0) {
      edge e = dominic_successor(G, pos, cur, t);
      dist[cur] = 0.0;
      pred[cur] = e;
      P.push(G.target(e));
      cur = G.target(e);
    }

    // check if we're trapped in a cycle, and if so mark all nodes
    // on the path as being at distance infinity from v
    node prev = P.pop();
    if (prev != t && dist[prev] == 0.0) {
      dist[prev] = HUGE_VAL;
      while (!P.empty())
	dist[P.pop()] = HUGE_VAL;
    }

    // compute the distances by working backwards in the path
    while (!P.empty()) {
      node cur = P.pop();
      dist[cur] = pos[cur].distance(pos[prev]) + dist[prev];
      prev = cur;
    }
  }
}







