//==========================================================================
// File: voronoi.c
// Description: An implementation of Voronoi routing
// Author: Pat Morin
// Created: Sun Mar 28 21:45:10 EST 1999
//==========================================================================
#include "voronoi.h"

#include <LEDA/plane_alg.h>
#include <LEDA/window.h>

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif // min

//==========================================================================
// Return true if the edge e is on the boundary of a triangular face
// of DT and false otherwise (i.e., if it is on the boundary of the
// convex hull).
// Complexity: O(1)
// Precondition: DT is a Delaunay triangulation
//==========================================================================
inline
bool triangular_face(graph &DT, node_array<point> &pos, edge e)
{
  return left_turn(pos[DT.source(e)], pos[DT.target(e)],
		   pos[DT.target(DT.cyclic_adj_succ(e))]);
}

//==========================================================================
// Return a segment representing the dual (Voronoi) edge of an edge in
// the Delaunay triangulation.
// Complexity: O(1)
// Precondition: DT is a Delaunay triangulation
//==========================================================================
inline
segment dual_edge(graph &DT, node_array<point> &pos, edge e)
{
  point p1;
  if (triangular_face(DT, pos, e)) {
    p1 = circle(pos[DT.source(e)], pos[DT.target(e)], 
		pos[DT.target(DT.cyclic_adj_succ(e))]).center();
  } else {
    vector v = pos[DT.target(e)] - pos[DT.source(e)];
    point m((pos[DT.source(e)].xcoord() + pos[DT.target(e)].xcoord()) / 2,
	    (pos[DT.source(e)].ycoord() + pos[DT.target(e)].ycoord()) / 2);
    p1 = m + 20*v.rotate90();
  }

  e = DT.reversal(e);
  point p2;
  if (triangular_face(DT, pos, e)) {
    p2 = circle(pos[DT.source(e)], pos[DT.target(e)], 
		pos[DT.target(DT.cyclic_adj_succ(e))]).center();
  } else {
    vector v = pos[DT.target(e)] - pos[DT.source(e)];
    point m((pos[DT.source(e)].xcoord() + pos[DT.target(e)].xcoord()) / 2,
	    (pos[DT.source(e)].ycoord() + pos[DT.target(e)].ycoord()) / 2);
    p2 = m + 20*v.rotate90();
  }
  return segment(p1, p2);
}

//==========================================================================
// Compute the Voronoi path from s to t and store it in P
// Complexity: O(n)
// Precondition: G is a Delaunay triangulation
//==========================================================================
void VORONOI_PATH(graph &G, node_array<point> &pos, 
		  node s, node t, list<node> &P)
{
  segment st(pos[s], pos[t]);
  P.clear();
  P.append(s);
  edge e = G.first_adj_edge(s);
  while (G.source(e) != t) {
    while (!dual_edge(G, pos, e).intersection(st))
      e = G.cyclic_adj_succ(e);
    e = G.cyclic_adj_succ(G.reversal(e));
    P.append(G.source(e));
  }
}

void exponential_search(node_array<point> &pos, list<node> &P1, 
			list<node> &P2, list<node> &P)
{
  list<node> tmpP;

  // compute the minimum length
  double len = min(pos[P1.front()].distance(pos[P1[P1.succ(P1.first())]]),
		   pos[P2.front()].distance(pos[P2[P2.succ(P2.first())]]));

  // do exponential search on the two paths
  stack<node> S;
  do {
    double d = 0.0;
    S.push(P1.pop());
    tmpP.append(S.top());
    while (!P1.empty() 
	   && d + pos[S.top()].distance(pos[P1.front()]) < 2*len) {
      d += pos[S.top()].distance(pos[P1.front()]);
      S.push(P1.pop());
      tmpP.append(S.top());
    }
    if (!P1.empty()) {
      d += pos[S.top()].distance(pos[P1.front()]);
      len = d;
      // P1.push(S.pop());
      while (!S.empty()) {
	tmpP.append(S.top());
	P1.push(S.pop());
      }
      d = 0.0;
      S.push(P2.pop());
      tmpP.append(S.top());
      while (!P2.empty() 
	     && d + pos[S.top()].distance(pos[P2.front()]) < 2*len) {
	d += pos[S.top()].distance(pos[P2.front()]);
	S.push(P2.pop());
	tmpP.append(S.top());
      }
      if (!P2.empty()) {
	d += pos[S.top()].distance(pos[P2.front()]);
	len = d;
	// P2.push(S.pop());
	while (!S.empty()) {
	  tmpP.append(S.top());
	  P2.push(S.pop());
	}
      }
    }
  } while (!P1.empty() && !P2.empty());
  P.conc(tmpP);
}

//==========================================================================
// Compute the parallel Voronoi path from s to t (which may actually
// be a walk rather than a path) and store it in P.
// the Voronoi site in t
// Complexity: O(n)
// Precondition: G is a Delaunay triangulation
//==========================================================================
void PARALLEL_VORONOI_PATH(graph &G, node_array<point> &pos, 
			   node s, node t, list<node> &P)
{
  list<node> Pd;
  VORONOI_PATH(G, pos, s, t, Pd);
  segment st(pos[s], pos[t]);
  node prev = Pd.pop();
  P.append(prev);
  while (!Pd.empty()) {
    node next = Pd.pop();
    if (!Pd.empty() 
	&& !right_turn(pos[s], pos[t], pos[prev])
	&& right_turn(pos[s], pos[t], pos[next])) {
      // compute the direct path that dips below st
      list<node> P1;
      P1.append(prev);
      P1.append(next);
      while (!Pd.empty() && right_turn(pos[s], pos[t], pos[next])) {
	next = Pd.pop();
	P1.append(next);
      }
      // compute the face path above st
      list<node> P2;
      P2.append(prev);
      edge e = G.first_adj_edge(prev);
      while (G.target(e) != next
	     && !right_turn(pos[s], pos[t], pos[G.target(e)])) 
	e = G.cyclic_adj_succ(e);
      while (G.source(e) != next) {
	while (G.target(e) != next
	       && right_turn(pos[s], pos[t], pos[G.target(e)]))
	  e = G.cyclic_adj_succ(e);
	e = G.cyclic_adj_succ(G.reversal(e));
	P2.append(G.source(e));
      }
      // find the shorter of the two using exponential search
      exponential_search(pos, P1, P2, P); 
    } else {
      P.append(next);
    }
    prev = next;
  }
  // clean up the path a bit
  for (list_item it = P.first(); it != nil; it = P.succ(it)) {
    while (P.succ(it) != nil && P[P.succ(it)] == P[it]) {
      P.del(P.succ(it));
    }
  }
}

//==========================================================================
// Find the Voronoi path between all pairs of vertices in G.
// Complexity: O(n^3)
// Precondition: G is a Delaunay triangulation
//========================================================================
void VORONOI_TEST(graph &G, node_array<point>& pos)
{
  window W(600+20, 600+20, "Voronoi path debugging");  
  W.button("continue",1);
  W.open(window::center,window::center);
  W.init(-.1, 1.2, -.1);

  node s, t;
  // find the voronoi path between every pair of sites
  forall_nodes (s, G) {
    t = G.succ_node(s);
    while (t != nil) {
      list<node> P;
      PARALLEL_VORONOI_PATH(G, pos, s, t, P);
      W.clear();
      W.set_node_width(2);
      W.set_line_width(1);
      edge e;
      forall_edges(e, G) {
	// W.draw_segment(pos[G.source(e)], pos[G.target(e)], red);
	// W.draw_filled_node(pos[G.source(e)], red);
	W.draw_segment(dual_edge(G, pos, e), blue);
      }
      W.draw_segment(pos[s], pos[t], green);
      W.draw_filled_node(pos[s], green);
      W.draw_filled_node(pos[t], blue);
      W.set_line_width(2);
      node prev = P.pop();
      while (!P.empty()) {
	W.draw_filled_node(pos[prev], black);
	// wait for input
	while (1 < 2) { 
	  double x, y;
	  int k = W.read_mouse(x, y);
	  if (k==1) break;
	}
	node next = P.pop();
	W.draw_filled_node(pos[prev], green);
	W.draw_segment(pos[prev], pos[next], orange);
	prev = next;
      }
      // wait for input
      while (1 < 2) { 
	double x, y;
	int k = W.read_mouse(x, y);
	if (k==1) break;
      }
      t = G.succ_node(t);
    }
  }
}





